/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.results;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import monalisa.Project;
import monalisa.data.pn.PInvariant;
import monalisa.data.pn.Place;

public final class PInvariants implements Result, Collection<PInvariant> {
    private static final long serialVersionUID = 8293263678484610772L;
    private final List<PInvariant> pinvariants;
    
    public PInvariants(List<PInvariant> pinvariants) {
        this.pinvariants = Collections.unmodifiableList(pinvariants);
    }
    
    @Override
    public void export(File path, Configuration config, Project project) throws IOException {
        PrintWriter printer = null;
        try {
            printer = new PrintWriter(path);
            
            String text;
            printer.println("# List of all place invaraints with the internal IDs of the species");
            for (PInvariant pinvariant : pinvariants) {
                text = String.format("%d.Place Invariant(%d): %s",pinvariant.id()+1, pinvariant.size(), placesToString(pinvariant,true));
                printer.println(text);
            }
            
            printer.println("\n\n# List of all place invaraints with the names of the species");
                        
            for (PInvariant pinvariant : pinvariants) {
                text = String.format("%d.Place Invariant(%d): %s",pinvariant.id()+1, pinvariant.size(), placesToString(pinvariant,false));
                printer.println(text);
            }
        }
        finally {
            if (printer != null)
                printer.close();
        }
    }
    
    @Override
    public String toString() {
        return pinvariants.toString();
    }
    
    private static String placesToString(PInvariant pinvariant, Boolean printId) {
        StringBuilder ret = new StringBuilder();
        boolean first = true;
        
        for (Place place : pinvariant) {
            if (first)
                first = false;
            else
                ret.append(" ");
            if (pinvariant.factor(place) != 1)
                ret.append(String.format("%d*", pinvariant.factor(place)));
            
            if(printId)
                ret.append(place.id()+1);
            else
                ret.append(((String) place.getProperty("name")).replace(" ", "_"));                
        }
        
        return ret.toString();
    }
    
    private static List<String> paragraphize(String text, String hangingIndent, int lineLength) {
        List<String> ret = new ArrayList<String>();
        int startPos = 0;
        String indent = "";
        
        while (text.length() - startPos > lineLength) {
            // Go forward to the theoretical end of the line and walk backwards
            // to the beginning of the current word.
            int prev = startPos;
            startPos += lineLength;
            while (!Character.isWhitespace(text.charAt(startPos)) && startPos >= prev)
                startPos--;
            // Failsafe if the line has no spaces:
            if (startPos == prev)
                startPos += lineLength; // Don't care: cut the word.
            
            ret.add(indent + text.substring(prev, startPos));
            startPos++; // Skip whitespace.
            
            // Adjust line length by hanging indent.
            lineLength -= hangingIndent.length();
            indent = hangingIndent;
        }
        
        // Add the dangling line.
        if (startPos < text.length())
            ret.add(indent + text.substring(startPos));
        
        return ret;
    }

    @Override
    public String filenameExtension() {
        return "inv";
    }
    
//    public Collection<PInvariant> nonTrivialTInvariants() {
//        List<PInvariant> nonTrivial = new ArrayList<PInvariant>();
//        for(PInvariant pinvariant : this) {
//            if(!pinvariant.isTrivial())
//                nonTrivial.add(pinvariant);
//        }
//        return nonTrivial;
//    }

    @Override
    public Iterator<PInvariant> iterator() {
        return pinvariants.iterator();
    }

    @Override
    public boolean add(PInvariant o) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean addAll(Collection<? extends PInvariant> c) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void clear() {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean contains(Object o) {
        return pinvariants.contains(o);
    }

    @Override
    public boolean containsAll(Collection<?> c) {
        return pinvariants.containsAll(c);
    }

    @Override
    public boolean isEmpty() {
        return pinvariants.isEmpty();
    }

    @Override
    public boolean remove(Object o) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean removeAll(Collection<?> c) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean retainAll(Collection<?> c) {
        throw new UnsupportedOperationException();
    }

    @Override
    public int size() {
        return pinvariants.size();
    }

    @Override
    public Object[] toArray() {
        return pinvariants.toArray();
    }

    @Override
    public <T> T[] toArray(T[] a) {
        return pinvariants.toArray(a);
    }

    public PInvariant getTInvariant(int i) {
        return pinvariants.get(i);
    }
}