/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.results;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import monalisa.Project;
import monalisa.data.pn.TInvariant;
import monalisa.data.pn.Transition;

public final class TInvariants implements Result, Collection<TInvariant> {
    private static final long serialVersionUID = 8293263678484610772L;
    private final List<TInvariant> tinvariants;
    
    public TInvariants(List<TInvariant> tinvariants) {
        this.tinvariants = Collections.unmodifiableList(tinvariants);
    }
    
    @Override
    public void export(File path, Configuration config, Project project) throws IOException {
        PrintWriter printer = null;
        try {
            printer = new PrintWriter(path);
                        
            String text;
            printer.println("# List of all elementary modes with the internal IDs of the reactions");
            for (TInvariant tinvariant : tinvariants) {
                text = String.format("%d.Elementary Mode(%d): %s", tinvariant.id()+1, tinvariant.size(), transitionsToString(tinvariant, true));                
                printer.println(text);
            }
            
            printer.println("\n\n# List of all elementary modes with the names of the reactions");
            
            for (TInvariant tinvariant : tinvariants) {
                text = String.format("%d.Elementary Mode(%d): %s", tinvariant.id()+1, tinvariant.size(), transitionsToString(tinvariant, false));                
                printer.println(text);
            }
        }
        finally {
            if (printer != null)
                printer.close();
        }
    }

    @Override
    public String toString() {
        return tinvariants.toString();
    }
    
    private static String transitionsToString(TInvariant tinvariant, Boolean printId) {
        StringBuilder ret = new StringBuilder();
        boolean first = true;
        
        for (Transition transition : tinvariant) {
            if (first)
                first = false;
            else
                ret.append(" ");
            if (tinvariant.factor(transition) != 1)
                ret.append(String.format("%d*", tinvariant.factor(transition)));

            if(printId)                
                ret.append(transition.id()+1);
            else
                ret.append(((String) transition.getProperty("name")).replace(" ", "_"));
        }
        
        return ret.toString();
    }
    
    private static List<String> paragraphize(String text, String hangingIndent, int lineLength) {
        List<String> ret = new ArrayList<String>();
        int startPos = 0;
        String indent = "";
        
        while (text.length() - startPos > lineLength) {
            // Go forward to the theoretical end of the line and walk backwards
            // to the beginning of the current word.
            int prev = startPos;
            startPos += lineLength;
            while (!Character.isWhitespace(text.charAt(startPos)) && startPos >= prev)
                startPos--;
            // Failsafe if the line has no spaces:
            if (startPos == prev)
                startPos += lineLength; // Don't care: cut the word.
            
            ret.add(indent + text.substring(prev, startPos));
            startPos++; // Skip whitespace.
            
            // Adjust line length by hanging indent.
            lineLength -= hangingIndent.length();
            indent = hangingIndent;
        }
        
        // Add the dangling line.
        if (startPos < text.length())
            ret.add(indent + text.substring(startPos));
        
        return ret;
    }

    @Override
    public String filenameExtension() {
        return "inv";
    }
    
    public Collection<TInvariant> nonTrivialTInvariants() {
        List<TInvariant> nonTrivial = new ArrayList<TInvariant>();
        for(TInvariant tinvariant : this) {
            if(!tinvariant.isTrivial())
                nonTrivial.add(tinvariant);
        }
        return nonTrivial;
    }

    @Override
    public Iterator<TInvariant> iterator() {
        return tinvariants.iterator();
    }

    @Override
    public boolean add(TInvariant o) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean addAll(Collection<? extends TInvariant> c) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void clear() {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean contains(Object o) {
        return tinvariants.contains(o);
    }

    @Override
    public boolean containsAll(Collection<?> c) {
        return tinvariants.containsAll(c);
    }

    @Override
    public boolean isEmpty() {
        return tinvariants.isEmpty();
    }

    @Override
    public boolean remove(Object o) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean removeAll(Collection<?> c) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean retainAll(Collection<?> c) {
        throw new UnsupportedOperationException();
    }

    @Override
    public int size() {
        return tinvariants.size();
    }

    @Override
    public Object[] toArray() {
        return tinvariants.toArray();
    }

    @Override
    public <T> T[] toArray(T[] a) {
        return tinvariants.toArray(a);
    }

    public TInvariant getTInvariant(int i) {
        return tinvariants.get(i);
    }

    public TInvariant getTInvariant(String name) {
        TInvariant inv = null;
        for(TInvariant t : tinvariants) {
            if(t.toString().equals(name)) {
                inv = t;
                break;
            }
        }

        return inv;
    }
}